/******************************************************************************
*
* project name:    TIGCC Tools Suite
* file name:       ttstart.c
* initial date:    14/08/2000
* author:          thomas.nussbaumer@gmx.net
* description:     start utility for compressed programs on the calculator
*
* $Id: ttstart.c,v 1.7 2000/11/18 16:00:18 Thomas Nussbaumer Exp $
*
******************************************************************************/

//----------------------------------------------------------------------------
// programs which should be startable with ttstart must be truncated by 
// ttstrip, than packed with ttpack and afterwards encapsulated with ttbin2oth 
// into an OTH variable with extension "ppg" (packed program)
//----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// you can also build a complete customized starter tool which doesn't need a 
// commandline argument!
//
// Suppose your ppg file is named myprog and you want the starter tool be 
// called mystart, then compile ttstart.c this way:
//
// tigcc -o mystart -O2 -DCUSTOM_NAME=myprog ttstart.c
//
//-----------------------------------------------------------------------------

#define OPTIMIZE_ROM_CALLS
#include <statline.h>
#include <alloc.h>
#include <vat.h>
#include <string.h>
#include <system.h>

#ifndef CUSTOM_NAME
#include <args.h>    
#endif
int _ti92plus;
int _ti89;

//#define TIME_MEASUREMENT

#define fatal(s) {ST_showHelp(s); return;}

#include "asm_upck.c"               // decompression routine
#include "../include/ttunpack.h"    // errorcodes definition
#include "../include/packhead.h"    // compressed header definition

//-----------------------------------------------------------------------------
// NOTE: the NOPs are very important due to a dirty hack in the TIGCCLIB with 
//       define RETURN_VALUE (they must be exactly after the call!!)
//-----------------------------------------------------------------------------
#define ASM_NOP_CALL(x) ({\
asm volatile("movem.l %d0-%d7/%a0-%a6,-(%sp)"); \
ASM_fastcall(x);\
asm volatile("nop;nop;nop;nop;");\
asm volatile("movem.l (%sp)+,%d0-%d7/%a0-%a6");})


//=============================================================================
// nothing special, just a huge main routine ...
//=============================================================================
void _main(void) {
    unsigned char*  src;
    unsigned char*  dest;
    unsigned short  length;
    char            tmpstr[50];
    char*           fname;
    SYM_ENTRY*      symptr;
    volatile HANDLE h;
#ifdef TIME_MEASUREMENT
    unsigned long  timercount;
#endif
#ifndef CUSTOM_NAME
    ESI            argptr;
    int            argtype;

    InitArgPtr(argptr);
    argtype = GetArgType(argptr);

    if (argtype != STR_TAG) fatal("use: ttstart(\"name\")");

    fname = GetStrnArg(argptr);
#else
    //-------------------------------------------------------------------------
    // if you can figure out why the next lines are that strange, then you are 
    // really good in C programming ...
    //-------------------------------------------------------------------------
    #define TOSTR(x) #x		
    #define XTOSTR(x) TOSTR(x)

    fname = XTOSTR(CUSTOM_NAME);
#endif    

    //-------------------------------------------------------------------------
    // setup TIOS name for variable and try to find it in the current folder ..
    //-------------------------------------------------------------------------
    tmpstr[0] = 0;
    strcpy(tmpstr+1,fname);
    if (!(symptr = DerefSym(SymFind(tmpstr+1+strlen(fname))))) {
        sprintf(tmpstr,"%s not found",fname);
        fatal(tmpstr);
    }

    //-------------------------------------------------------------------------
    // IMPORTANT: only use handle (symptr may become invalid !!!)
    //-------------------------------------------------------------------------
    h = symptr->handle;

    //-------------------------------------------------------------------------
    // if file is already locked, don't lock it again ...
    //-------------------------------------------------------------------------
    if (HeapGetLock(h)) {
        src = HeapDeref(h);
        h   = 0;
    }
    else {
        src = HLock(h);
    }
    if (!src) fatal("locking failed");
   
    length = *(unsigned int*)src+3;
    dest   = src+length-2;
    
    //-------------------------------------------------------------------------
    // skip "length of variable" bytes
    //-------------------------------------------------------------------------
    src+=2;  

    //-------------------------------------------------------------------------
    // check extension and if its a packed file
    //-------------------------------------------------------------------------
    if (*dest     != 0xF8 ||  
        *(dest-4) != 'p'  ||
        *(dest-3) != 'p'  || 
        *(dest-2) != 'g'  ||
        !IsPacked(((PackedHeader*)src))) 
    {
        if (h) HeapUnlock(h);
        fatal("no ppg file");
    }

    //-------------------------------------------------------------------------
    // get the original size
    //-------------------------------------------------------------------------
    length = GetUnPackedSize(((PackedHeader*)src)); 

    //-------------------------------------------------------------------------
    // allocate memory for original
    //-------------------------------------------------------------------------
    if (!(dest=malloc(length))) {     
        if (h) HeapUnlock(h);
        fatal("out of memory"); 
    }

    ST_showHelp("decompressing ...");

#ifdef TIME_MEASUREMENT
    //-------------------------------------------------------------------------
    // start timer for clocks and enter the main loop
    //-------------------------------------------------------------------------
    OSFreeTimer(USER_TIMER);
    OSRegisterTimer(USER_TIMER,1000);
    OSTimerRestart(USER_TIMER);
#endif

#ifdef TIME_MEASUREMENT
    {
        int retval = _tt_Decompress(src,dest);

        timercount = OSTimerCurVal(USER_TIMER);
        //---------------------------------------------------------------------
        // stop timer again, turn grayscales off, restore
        // old interrupt handlers again and free memory
        //---------------------------------------------------------------------
        OSFreeTimer(USER_TIMER);
        sprintf(tmpstr,"%lu tics",(1000-timercount)*50);
        ST_showHelp(tmpstr);
        ngetchx();

        if (retval) {
            char s[100];
            sprintf(s,"e=%d (%u)",retval,length);
            free(dest);
            if (h) HeapUnlock(h);
            fatal(s);
        }
    }
#else 
    if (_tt_Decompress(src,dest)) {
        //---------------------------------------------------------------------
        // "houston we got problems ..."
        //---------------------------------------------------------------------
        free(dest);
        if (h) HeapUnlock(h);
        fatal("failed");
    }        
#endif

    ST_eraseHelp();
    //-------------------------------------------------------------------------
    // we can unlock the original already ...
    //-------------------------------------------------------------------------
    if (h) HeapUnlock(h);
    
    //-------------------------------------------------------------------------
    // relocate program and start it ...
    //-------------------------------------------------------------------------
    enter_ghost_space();  
    EX_patch(dest+0x40002,dest+length-1);
    ASM_NOP_CALL(dest+0x40002);
    free(dest);
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: ttstart.c,v $
// Revision 1.7  2000/11/18 16:00:18  Thomas Nussbaumer
// enter_ghost_space() added
//
// Revision 1.6  2000/10/01 14:56:52  Thomas Nussbaumer
// support for custom start utility build added
//
// Revision 1.5  2000/08/30 20:01:55  Thomas Nussbaumer
// using now assembler optimized version of unpack routine
//
// Revision 1.4  2000/08/30 19:44:24  Thomas Nussbaumer
// TABs replaced by spaces and trailing spaces removed
//
// Revision 1.3  2000/08/20 15:28:22  Thomas Nussbaumer
// including of unpack.c done now before including ttunpack.h
//
// Revision 1.2  2000/08/16 22:29:05  Thomas Nussbaumer
// (1) based now on shared file unpack.c
// (2) argument can be specified on commandline
// (3) define and code for time measurement integrated
//
// Revision 1.1  2000/08/14 22:48:38  Thomas Nussbaumer
// initial version
//
//
